﻿using System.Linq;
using System.Runtime.Remoting.Contexts;
using Singular.Dynamics;
using Singular.Helpers;
using Singular.Managers;
using Singular.Settings;
using Styx;

using Styx.CommonBot;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.TreeSharp;
using Rest = Singular.Helpers.Rest;
using System;

using CommonBehaviors.Actions;

namespace Singular.ClassSpecific.Paladin
{
    public static class Holy
    {
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static PaladinSettings PaladinSettings { get { return SingularSettings.Instance.Paladin(); } }
        public static bool HasTalent(PaladinTalents tal) { return TalentManager.IsSelected((int)tal); }

        const int PriEmergencyBase = 500;
        const int PriHighBase = 400;
        const int PriAoeBase = 300;
        const int PriSingleBase = 100;
        const int PriLowBase = 0;

        private static int HealthToPriority(int nHealth)
        {
            return nHealth == 0 ? 0 : 200 - nHealth;
        }

        [Behavior(BehaviorType.Rest, WoWClass.Paladin, WoWSpec.PaladinHoly)]
        public static Composite CreatePaladinHolyRest()
        {
            return new PrioritySelector(
                Spell.WaitForCast(),
                new Decorator(
                    ret => !Spell.IsGlobalCooldown(),
                    new PrioritySelector(
                        // Heal self before resting. There is no need to eat while we have 100% mana
                        CreatePaladinHealBehavior(true),
                        // Rest up damnit! Do this first, so we make sure we're fully rested.
                        Rest.CreateDefaultRestBehaviour( null, "Redemption"),
                        // Make sure we're healing OOC too!
                        CreatePaladinHealBehavior(false, false)
                        )
                    )
                );
        }
        [Behavior(BehaviorType.Heal, WoWClass.Paladin, WoWSpec.PaladinHoly)]
        public static Composite CreatePaladinHolyHealBehavior()
        {
            return new PrioritySelector(
                Spell.BuffSelf( "Devotion Aura", req => Me.Silenced ),
                CreatePaladinHealBehavior(),
                new Decorator(
                    req => !Unit.NearbyGroupMembers.Any(m => m.IsAlive && !m.IsMe),
                    new PrioritySelector(
                        Spell.Cast("Lay on Hands",
                            mov => false,
                            on => Me,
                            req => Me.PredictedHealthPercent(includeMyHeals: true) <= PaladinSettings.SelfLayOnHandsHealth),
                        Common.CreateWordOfGloryBehavior(on => Me),
                        Spell.Cast("Flash of Light",
                            mov => true,
                            on => Me,
                            req => Me.PredictedHealthPercent(includeMyHeals: true) <= PaladinSettings.SelfFlashOfLightHealth,
                            cancel => Me.HealthPercent > PaladinSettings.SelfFlashOfLightHealth)
                        )
                    )
                );
        }
        
        [Behavior(BehaviorType.Combat, WoWClass.Paladin, WoWSpec.PaladinHoly)]
        public static Composite CreatePaladinHolyCombatBehavior()
        {
            return new PrioritySelector(
                new Decorator(
                    ret => !Unit.NearbyGroupMembers.Any(m => m.IsAlive && !m.IsMe),
                    new PrioritySelector(
                        Helpers.Common.EnsureReadyToAttackFromMelee(),
                        Spell.WaitForCastOrChannel(),

                        new Decorator( 
                            ret => !Spell.IsGlobalCooldown() && Me.GotTarget,
                            new PrioritySelector(
                                Helpers.Common.CreateAutoAttack(true),
                                Helpers.Common.CreateInterruptBehavior(),

                                Common.CreatePaladinPullMore(),

                                Common.CreatePaladinBlindingLightBehavior(),

                                Spell.Cast("Hammer of Justice", ret => PaladinSettings.StunMobsWhileSolo && SingularRoutine.CurrentWoWContext == WoWContext.Normal),
                                Spell.Buff("Judgment"),
                                Spell.Cast("Hammer of Wrath"),
                                Spell.Cast("Holy Shock"),
                                Spell.Cast("Crusader Strike"),
                                Spell.Cast("Denounce")
                                )
                            ),

                        Movement.CreateMoveToMeleeBehavior(true)
                        )
                    )
                );
        }
        
        [Behavior(BehaviorType.Pull, WoWClass.Paladin, WoWSpec.PaladinHoly)]
        public static Composite CreatePaladinHolyPullBehavior()
        {
            return new PrioritySelector(
                new Decorator(
                    ret => !StyxWoW.Me.GroupInfo.IsInParty && !StyxWoW.Me.GroupInfo.IsInRaid,
                    new PrioritySelector(
                        Helpers.Common.EnsureReadyToAttackFromMelee(),
                        Spell.Cast("Judgment"),
                        Helpers.Common.CreateAutoAttack(true),
                        Movement.CreateMoveToMeleeBehavior(true)
                        ))
                );
        }

        internal static Composite CreatePaladinHealBehavior()
        {
            return CreatePaladinHealBehavior(false, true);
        }

        internal static Composite CreatePaladinHealBehavior(bool selfOnly)
        {
            return CreatePaladinHealBehavior(selfOnly, false);
        }

        private static WoWUnit _healTarget;

        internal static Composite CreatePaladinHealBehavior(bool selfOnly, bool moveInRange)
        {
            if (SingularRoutine.CurrentWoWContext == WoWContext.Normal)
                return new ActionAlwaysFail();

            HealerManager.NeedHealTargeting = true;
            PrioritizedBehaviorList behavs = new PrioritizedBehaviorList();
            int cancelHeal = (int)Math.Max(SingularSettings.Instance.IgnoreHealTargetsAboveHealth, Math.Max(PaladinSettings.FlashOfLightHealth, PaladinSettings.HolyLightHealth));

            Logger.WriteDebugInBehaviorCreate("Paladin Healing: will cancel cast of direct heal if health reaches {0:F1}%", cancelHeal);

            #region Cleanse

            if (SingularSettings.Instance.DispelDebuffs != RelativePriority.None)
            {
                int dispelPriority = (SingularSettings.Instance.DispelDebuffs == RelativePriority.HighPriority) ? 999 : -999;
                behavs.AddBehavior(dispelPriority, "Cleanse", "Cleanse", Dispelling.CreateDispelBehavior());
            }

            #endregion

            #region Beacon

            // Beacon Tank
            behavs.AddBehavior(805, "Beacon of Light", "Beacon of Light",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Buff(
                            "Beacon of Light",
                            on =>
                            {
                                if (PaladinSettings.BeaconOfLightTankOnly || Group.Tanks.Any(u => u.HealthPercent <= PaladinSettings.BeaconOfLightTankHealth) && GetTimeSinceLastBeaconTimer() >= 5000)
                                {
                                    //Logger.WriteDebug("TimeSinceLastBeacon: {0} ms", GetTimeSinceLastBeaconTimer());
                                    BeaconTimer.Restart();
                                    return GetBestTankTarget(PaladinSettings.BeaconOfLightTankOnly || Unit.FriendlyPlayers(60).Any(u => !u.HasMyAura("Beacon of Light") && u.IsInMyPartyOrRaid) ? 100 : PaladinSettings.BeaconOfLightTankHealth, "Beacon of Light");
                                }
                                else
                                {
                                    return (WoWUnit)on;
                                }
                            },
                            req => true
                            )
                        )
                    );

            #endregion

            #region Sacred Shield

            // Sacred Shield
            behavs.AddBehavior(795, "Sacred Shield @ players", "Sacred Shield",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Cast(
                            "Sacred Shield",
                            on =>
                            {
                                // Sacred Shield stacks duration, so use it on tank if SS Time Left <= 8000
                                if (Group.Tanks.Any(u => u.HasMyAura(148039) && u.GetAuraTimeLeft(148039).TotalMilliseconds <= 9000) && PaladinSettings.SacredShieldTankPriority)
                                {
                                    return Group.Tanks.Where(u => u.HasMyAura(148039) && u.GetAuraTimeLeft(148039).TotalMilliseconds <= 9000).OrderBy(u => u.HealthPercent).FirstOrDefault();
                                }

                                bool tankNeedsSS = false;
                                float hp = PaladinSettings.SacredShieldHealth;

                                if (Group.Tanks.Any(u => !u.HasMyAura(148039)) && PaladinSettings.SacredShieldTankPriority)
                                {
                                    hp = 100;
                                    tankNeedsSS = true;
                                }

                                return GetBestTargetBuff(hp, "Sacred Shield", tankNeedsSS);
                            },
                            req => true,
                            cancel => false
                        )
                    )
                );

            #endregion

            #region Save the Group

            // Lay on Hands anyone
            if (PaladinSettings.LayOnHandsHealth != 0)
                behavs.AddBehavior(800, "Lay on Hands @ " + PaladinSettings.LayOnHandsHealth, "Lay on Hands",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        new PrioritySelector(
                            ctx => HealerManager.FindHighestPriorityTarget(),
                            Spell.Cast(
                                "Lay on Hands",
                                mov => true,
                                on => (WoWUnit)on,
                                req => ((WoWUnit)req).HealthPercent < PaladinSettings.LayOnHandsHealth
                                )
                            )
                        )
                    );

            // Hand of Sac anyone
            if (PaladinSettings.HandOfSacrificeHealth != 0)
            {
                behavs.AddBehavior(HealthToPriority(PaladinSettings.HandOfSacrificeHealth) + PriEmergencyBase, "Hand of Sacrifice @ " + PaladinSettings.HandOfSacrificeHealth, "Hand of Sacrifice",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Buff(
                            "Hand of Sacrifice",
                            on => (WoWUnit)on,
                            req => ((WoWUnit)req).HealthPercent < PaladinSettings.HandOfSacrificeHealth && ((WoWUnit)req).Guid != StyxWoW.Me.Guid && ((WoWUnit)req).ToPlayer().IsInMyPartyOrRaid
                            )
                        )
                    );
            }

            // Word of Glory Emergency Heal; Make sure this takes higher priority than LoD (We do not have EF, so we should be spamming LoD. Only use WoG if a target needs a quick heal)
            if (SpellManager.HasSpell("Word of Glory"))
            {
                behavs.AddBehavior(797, "Word of Glory @ players", "Word of Glory",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                            Spell.Cast(
                                "Word of Glory",
                                on => (WoWUnit)on,
                                req => StyxWoW.Me.CurrentHolyPower >= (uint)PaladinSettings.WordOfGloryHP && ((WoWUnit)req).PredictedHealthPercent() <= PaladinSettings.WordOfGloryEmergencyHealth,
                                cancel => false
                            )
                        )
                    );
            }

            // Flash of Light Emergency Heal Tank
            // Make priority a little higher than EF since it now has cast time
            behavs.AddBehavior(796, "Flash of Light Tank Emergency @ " + PaladinSettings.FlashOfLightTankHealth, "Flash of Light",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                    Spell.Cast(
                        "Flash of Light",
                        on => Group.Tanks.Where(u => u.HealthPercent <= PaladinSettings.FlashOfLightTankHealth && u.IsAlive).OrderBy(u => u.HealthPercent).FirstOrDefault(),
                        req => true
                        )
                    )
                );

            // Flash of Light Emergency Heal Anyone
            // Make priority a little higher than EF since it now has cast time
            behavs.AddBehavior(795, "Flash of Light Anyone Emergency @ " + PaladinSettings.FlashOfLightAnyoneHealth, "Flash of Light",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                    Spell.Cast(
                        "Flash of Light",
                        on => HealerManager.Instance.TargetList.Where(u => u.IsAlive && u.HealthPercent <= PaladinSettings.FlashOfLightAnyoneHealth).OrderBy(u => u.HealthPercent).FirstOrDefault(),
                        req => true
                        )
                    )
                );

            #endregion

            #region Light of Dawn

            // Light of Dawn
            if (PaladinSettings.LightOfDawnHealth != 0 && PaladinSettings.LightOfDawnCount != 0)
                behavs.AddBehavior(HealthToPriority(PaladinSettings.HolyRadianceHealth) + PriAoeBase + 3, "Light of Dawn @ " + PaladinSettings.LightOfDawnHealth + "% MinCount: " + PaladinSettings.LightOfDawnCount, "Light of Dawn",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                            Spell.Cast(
                                "Light of Dawn",
                                on => Clusters.GetBestUnitForCluster(Unit.NearbyFriendlyPlayers.Where(u => u.HealthPercent <= PaladinSettings.LightOfDawnHealth && u.IsAlive).Select(p => p.ToUnit()), ClusterType.Radius, 30f),
                                req => (StyxWoW.Me.CurrentHolyPower >= (uint)PaladinSettings.LightOfDawnHP || StyxWoW.Me.HasAura("Divine Purpose") || (SpellManager.HasSpell("Sacred Shield") && StyxWoW.Me.CurrentHolyPower == 5)) && HealerManager.Instance.TargetList.Count(p => p.IsAlive && p.HealthPercent <= PaladinSettings.LightOfDawnHealth) >= PaladinSettings.LightOfDawnCount,
                                cancel => StyxWoW.Me.CurrentHolyPower < (uint)PaladinSettings.LightOfDawnHP && !StyxWoW.Me.HasAura("Divine Purpose")
                                )
                            )
                        );

            #endregion

            #region Eternal Flame / Word of Glory

            // Eternal Flame
            if (SpellManager.HasSpell("Eternal Flame"))
            {
                behavs.AddBehavior(795, "Eternal Flame @ players", "Eternal Flame",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                            Spell.Cast(
                                "Eternal Flame",
                                on =>
                                {
                                    bool tankNeedsEF = false;
                                    float hp = PaladinSettings.EternalFlameHealth;
                                    int HoPo = PaladinSettings.EternalFlameHolyPower;

                                    if (Group.Tanks.Any(u => !u.HasMyAura("Eternal Flame")) && PaladinSettings.KeepEternalFlameUp)
                                    {
                                        hp = 100;
                                        tankNeedsEF = true;
                                    }
                                    else if (HoPo < 3)
                                    {
                                        hp = 100;
                                    }

                                    return GetBestTargetBuff(hp, "Eternal Flame", tankNeedsEF);
                                },
                                req =>
                                {
                                    bool tank = Group.Tanks.Any(u => !u.HasMyAura("Eternal Flame"));
                                    int HoPo = tank ? PaladinSettings.EternalFlameHolyPowerTank : PaladinSettings.EternalFlameHolyPower;

                                    return (StyxWoW.Me.CurrentHolyPower >= HoPo || StyxWoW.Me.HasAura("Divine Purpose")) && !((WoWUnit)req).HasMyAura("Eternal Flame");
                                },
                                cancel =>
                                {
                                    bool tank = Group.Tanks.Any(u => !u.HasMyAura("Eternal Flame"));
                                    int HoPo = tank ? PaladinSettings.EternalFlameHolyPowerTank : PaladinSettings.EternalFlameHolyPower;

                                    return StyxWoW.Me.CurrentHolyPower < HoPo && !StyxWoW.Me.HasAura("Divine Purpose");
                                }
                            )
                        )
                    );
            }

            // Word of Glory
            if (SpellManager.HasSpell("Word of Glory"))
            {
                behavs.AddBehavior(HealthToPriority(PaladinSettings.WordOfGloryHealth) + PriSingleBase, "Word of Glory @ players", "Word of Glory",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                            Spell.Cast(
                                "Word of Glory",
                                on => (WoWUnit)on,
                                req => (StyxWoW.Me.CurrentHolyPower >= (uint)PaladinSettings.WordOfGloryHP || StyxWoW.Me.HasAura("Divine Purpose")) && ((WoWUnit)req).PredictedHealthPercent() <= PaladinSettings.WordOfGloryHealth,
                                cancel => (StyxWoW.Me.CurrentHolyPower < (uint)PaladinSettings.WordOfGloryHP && !StyxWoW.Me.HasAura("Divine Purpose"))
                            )
                        )
                    );
            }

            #endregion

            #region Holy Shock
            // Holy Shock Daybreak Aoe
            behavs.AddBehavior(HealthToPriority(PaladinSettings.HolyRadianceHealth) + PriAoeBase + 2, "HolyShock Daybreak Aoe @ " + PaladinSettings.HolyShockHealth, "Holy Shock",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Cast(
                            "Holy Shock",
                            on => Clusters.GetBestUnitForCluster(Unit.NearbyFriendlyPlayers.Where(u => u.HealthPercent <= PaladinSettings.HolyShockHealth && u.IsAlive).Select(p => p.ToUnit()), ClusterType.Radius, 10f),
                            req => StyxWoW.Me.HasAura("Daybreak")
                            )
                        )
                    );

            // Holy Shock anything
            behavs.AddBehavior(HealthToPriority(PaladinSettings.HolyRadianceHealth) + PriAoeBase + 1, "HolyShock @ " + PaladinSettings.HolyShockHealth, "Holy Shock",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Cast(
                            "Holy Shock",
                            on => HealerManager.Instance.TargetList.Where(u => u.HealthPercent <= PaladinSettings.HolyShockHealth && u.IsAlive).OrderBy(u => u.HealthPercent).FirstOrDefault(),
                            req => true
                            )
                        )
                    );

            #endregion

            #region Aoe Heals

            // Holy Radiance
            if (PaladinSettings.HolyRadianceHealth != 0)
                behavs.AddBehavior(HealthToPriority(PaladinSettings.HolyRadianceHealth) + PriAoeBase, "HolyRadiance @ " + PaladinSettings.HolyRadianceHealth + "% MinCount: " + PaladinSettings.HolyRadianceCount, "Holy Radiance",
                    new Decorator(
                        ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                            Spell.Cast(
                                "Holy Radiance",
                                on => Clusters.GetBestUnitForCluster(Unit.NearbyFriendlyPlayers.Where(u => u.HealthPercent <= PaladinSettings.HolyRadianceHealth && u.IsAlive).Select(p => p.ToUnit()), ClusterType.Radius, 10f),
                                req => HealerManager.Instance.TargetList.Count(p => p.IsAlive && p.HealthPercent <= PaladinSettings.HolyRadianceHealth) >= PaladinSettings.HolyRadianceCount
                                )
                            )
                        );

            #endregion

            #region Single Heals

            // Holy Light
            behavs.AddBehavior(HealthToPriority(PaladinSettings.HolyLightHealth) + PriSingleBase, "HolyLight @ " + PaladinSettings.HolyLightHealth, "Holy Light",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Cast(
                            "Holy Light",
                            on => (WoWUnit)on,
                            req => ((WoWUnit)req).PredictedHealthPercent() <= PaladinSettings.HolyLightHealth,
                            cancel => ((WoWUnit)cancel).HealthPercent >= cancelHeal
                            )
                        )
                    );

            // Flash of Light
            behavs.AddBehavior(HealthToPriority(PaladinSettings.FlashOfLightHealth) + PriSingleBase, "FlashofLight @ " + PaladinSettings.FlashOfLightHealth, "Flash of Light",
                new Decorator(
                    ret => StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid,
                        Spell.Cast(
                            "Flash of Light",
                            on => (WoWUnit)on,
                            req => ((WoWUnit)req).PredictedHealthPercent() <= PaladinSettings.FlashOfLightHealth,
                            cancel => ((WoWUnit)cancel).HealthPercent >= cancelHeal
                            )
                        )
                    );

            #endregion

            #region Crusader Strike

            // Crusader Strike HP Generator
            behavs.AddBehavior(50 + PriSingleBase, "Crusader Strike " + PaladinSettings.UseCrusaderStrike, "Crusader Strike",
                new Decorator(
                    req => PaladinSettings.UseCrusaderStrike && StyxWoW.Me.Combat && Unit.ValidUnit(StyxWoW.Me.CurrentTarget) && !Unit.IsCrowdControlled(StyxWoW.Me.CurrentTarget),
                        Spell.Cast(
                            "Crusader Strike",
                            on => StyxWoW.Me.CurrentTarget,
                            req => StyxWoW.Me.CurrentHolyPower != 5
                            )
                        )
                    );

            #endregion

            #region Lowest Priority

            behavs.OrderBehaviors();

            if (selfOnly == false && Singular.Dynamics.CompositeBuilder.CurrentBehaviorType == BehaviorType.Heal)
                behavs.ListBehaviors();

            return new PrioritySelector(
                ctx => selfOnly ? StyxWoW.Me : HealerManager.FindLowestHealthTarget(), // HealerManager.Instance.FirstUnit,

                Spell.WaitForCastOrChannel(),

                new Decorator(
                    ret => !Spell.IsGlobalCooldown() && ret != null,
                    behavs.GenerateBehaviorTree()
                    ),

                new Decorator(
                    ret => moveInRange,
                    Movement.CreateMoveToUnitBehavior(
                        ret => Battlegrounds.IsInsideBattleground ? (WoWUnit)ret : Group.Tanks.Where(a => a.IsAlive).OrderBy(a => a.Distance).FirstOrDefault(),
                        35f
                        )
                    )
                );

            #endregion
        }

        private static WoWUnit GetBestTargetBuff(float health, string spellName, bool tanks)
        {
            WoWUnit bestTarget;

            if (tanks)
            {
                var tankList = Group.Tanks
                .Where(u => !u.HasMyAura(spellName) && u.IsAlive && u.HealthPercent <= health && u.SpellDistance() < 40 && u.InLineOfSpellSight)
                .OrderBy(u => u.HealthPercent)
                .ToList();

                bestTarget = (from u in tankList
                              orderby Unit.UnfriendlyUnitsNearTarget(u, 20).Where(a => a.CurrentTargetGuid == u.Guid).Count()
                              select u).FirstOrDefault();
            }
            else
            {
                bestTarget = Unit.NearbyGroupMembers
                    .Where(u => !u.HasMyAura(spellName) && u.IsAlive && u.HealthPercent <= health && u.SpellDistance() < 40 && u.InLineOfSpellSight)
                    .OrderBy(u => u.HealthPercent)
                    .FirstOrDefault();
            }

            if (bestTarget == null)
            {
                bestTarget = Unit.NearbyGroupMembers
                    .Where(u => !u.HasMyAura(spellName) && u.IsAlive && u.SpellDistance() < 40 && u.InLineOfSpellSight)
                    .OrderBy(u => u.HealthPercent)
                    .FirstOrDefault();
            }

            return bestTarget;
        }

        public static WoWUnit GetBestTankTarget(float health, string spellName)
        {
            var tankList = Group.Tanks
                .Where(u => !u.HasMyAura(spellName) && u.IsAlive && u.Combat && u.HealthPercent <= health && u.DistanceSqr < 40 * 40 && u.InLineOfSpellSight)
                .OrderBy(u => u.HealthPercent)
                .ToList();

            var bestTank = (from u in tankList
                            orderby Unit.UnfriendlyUnitsNearTarget(u, 20).Where(a => a.CurrentTargetGuid == u.Guid).Count() descending
                            select u).FirstOrDefault();

            if (bestTank == null)
            {
                bestTank = Group.Tanks
                    .Where(u => !u.HasMyAura(spellName) && u.IsAlive && u.Combat && u.HealthPercent <= health && u.DistanceSqr < 40 * 40 && u.InLineOfSpellSight)
                    .OrderBy(u => u.HealthPercent)
                    .FirstOrDefault();

                if (SingularSettings.Debug)
                    Logger.WriteDebug("GetBestTankTarget: Target == null. Found Tank {0} @ {1:F1}%", bestTank.SafeName(), bestTank.HealthPercent);
            }
            else
            {
                if (SingularSettings.Debug)
                    Logger.WriteDebug("GetBestTankTarget: found Tank {0} @ {1:F1}%, UnitsTargetingTank={2}", bestTank.SafeName(), bestTank.HealthPercent, Unit.UnfriendlyUnitsNearTarget(bestTank, 20).Where(a => a.CurrentTargetGuid == bestTank.Guid).Count());
            }

            return bestTank;
        }


        public static System.Diagnostics.Stopwatch BeaconTimer = new System.Diagnostics.Stopwatch();
        public static long GetTimeSinceLastBeaconTimer()
        {
            if (!BeaconTimer.IsRunning)
                BeaconTimer.Start();

            return BeaconTimer.ElapsedMilliseconds;
        }
    }
}
