﻿using System.Linq;
using Singular.Dynamics;
using Singular.Helpers;
using Singular.Managers;
using Singular.Settings;
using Styx;

using Styx.CommonBot;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

using Styx.TreeSharp;
using Action = Styx.TreeSharp.Action;
using Rest = Singular.Helpers.Rest;
using System.Drawing;
using System;

namespace Singular.ClassSpecific.Paladin
{
    public class Protection
    {

        #region Properties & Fields

        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static PaladinSettings PaladinSettings { get { return SingularSettings.Instance.Paladin(); } }

        private static int _aoeCount;

        private const int Prot_T16_ITEM_SET_ID = 1188;

        private static int NumTier16Pieces
        {
            get
            {
                return StyxWoW.Me.CarriedItems.Count(i => i.ItemInfo.ItemSetId == Prot_T16_ITEM_SET_ID);
            }
        }

        private static bool Has4PieceTier16Bonus { get { return NumTier16Pieces >= 4; } }

        #endregion

        [Behavior(BehaviorType.Rest, WoWClass.Paladin, WoWSpec.PaladinProtection)]
        public static Composite CreateProtectionRest()
        {
            return new PrioritySelector(
                // Rest up damnit! Do this first, so we make sure we're fully rested.
                Rest.CreateDefaultRestBehaviour( "Flash of Light", "Redemption")
                );
        }


        [Behavior(BehaviorType.Pull, WoWClass.Paladin, WoWSpec.PaladinProtection)]
        public static Composite CreatePaladinProtectionPull()
        {
            return new PrioritySelector(
                Helpers.Common.EnsureReadyToAttackFromMelee(),
                new Decorator(
                    ret => !Spell.IsGlobalCooldown(),
                    new PrioritySelector(
                        Helpers.Common.CreateAutoAttack(true),
                        Spell.BuffSelf("Sacred Shield", ret => !Me.HasMyAura(148039) || Me.GetAuraTimeLeft(148039).TotalMilliseconds <= 4000),
                        Spell.Cast("Judgment"),
                        Spell.Cast("Avenger's Shield"),
                        Spell.Cast("Reckoning", ret => !Me.CurrentTarget.IsPlayer)
                        )
                    ),

                Movement.CreateMoveToMeleeBehavior(true)
                );
        }

        [Behavior(BehaviorType.CombatBuffs, WoWClass.Paladin, WoWSpec.PaladinProtection)]
        public static Composite CreatePaladinProtectionCombatBuffs()
        {
            return new PrioritySelector(
                Spell.BuffSelf("Devotion Aura", req => Me.Silenced),

                // Seal twisting. If our mana gets stupid low, just throw on insight to get some mana back quickly, then put our main seal back on.
                // This is Seal of Truth once we get it, Righteousness when we dont.
                Common.CreatePaladinSealBehavior(),

                new Decorator(
                    req => !Unit.IsTrivial(Me.CurrentTarget),
                    new PrioritySelector(

                        Spell.WaitForCastOrChannel(),

                        // Defensive
                        Spell.BuffSelf("Hand of Freedom",
                            ret => Me.HasAuraWithMechanic(WoWSpellMechanic.Dazed,
                                                                    WoWSpellMechanic.Disoriented,
                                                                    WoWSpellMechanic.Frozen,
                                                                    WoWSpellMechanic.Incapacitated,
                                                                    WoWSpellMechanic.Rooted,
                                                                    WoWSpellMechanic.Slowed,
                                                                    WoWSpellMechanic.Snared)),

                        Spell.BuffSelf("Sacred Shield", ret => !Me.HasMyAura(148039) || Me.GetAuraTimeLeft(148039).TotalMilliseconds <= 4000),

                        Spell.BuffSelf("Divine Shield",
                            ret => Me.CurrentMap.IsBattleground && Me.HealthPercent <= 20 && !Me.HasAura("Forbearance")),

                        Spell.BuffSelf(
                            "Divine Protection",
                            ret => Me.HealthPercent <= PaladinSettings.DivineProtectionHealthProt),

                        Spell.BuffSelf(
                            "Guardian of Ancient Kings",
                            ret => Me.HealthPercent <= PaladinSettings.GoAKHealth),

                        Spell.BuffSelf(
                            "Ardent Defender",
                            ret => Me.HealthPercent <= PaladinSettings.ArdentDefenderHealth)
                        )
                    ),

                // Heal up after Defensive CDs used if needed
                Spell.BuffSelf( "Lay on Hands",
                    ret => Me.HealthPercent <= PaladinSettings.SelfLayOnHandsHealth && !Me.HasAura("Forbearance")),

                Common.CreateWordOfGloryBehavior(on => Me),

                Spell.Cast("Flash of Light",
                    mov => false,
                    on => Me,
                    req => SingularRoutine.CurrentWoWContext != WoWContext.Instances && Me.PredictedHealthPercent(includeMyHeals: true) <= PaladinSettings.SelfFlashOfLightHealth,
                    cancel => Me.HealthPercent > 90),

                // now any Offensive CDs
                new Decorator(
                    req => !Unit.IsTrivial(Me.CurrentTarget),
                    new PrioritySelector(

                        Spell.BuffSelf("Avenging Wrath", 
                            ret => (Me.GotTarget && Me.CurrentTarget.IsWithinMeleeRange && (Me.CurrentTarget.TimeToDeath() > 25) || _aoeCount > 1))

                        )
                    )
                );
        }

        [Behavior(BehaviorType.Combat, WoWClass.Paladin, WoWSpec.PaladinProtection, WoWContext.Normal | WoWContext.Battlegrounds)]
        [Behavior(BehaviorType.Combat, WoWClass.Paladin, WoWSpec.PaladinProtection, WoWContext.Instances)]
        public static Composite CreateProtectionCombat()
        {
            TankManager.NeedTankTargeting = (SingularRoutine.CurrentWoWContext == WoWContext.Instances);

            return new PrioritySelector(
                Helpers.Common.EnsureReadyToAttackFromMelee(),
                Spell.WaitForCastOrChannel(),

                new Decorator(
                    ret => !Spell.IsGlobalCooldown() && Unit.ValidUnit(StyxWoW.Me.CurrentTarget),
                    new PrioritySelector(

                        new Action( r => {
                            // Paladin AOE count should be those near paladin (consecrate, holy wrath) and those near target (avenger's shield)
                            _aoeCount = TankManager.Instance.TargetList.Count(u => u.SpellDistance() < 10 || u.Location.Distance(Me.CurrentTarget.Location) < 10);
                            return RunStatus.Failure;
                            }),

                        CreateProtDiagnosticOutputBehavior(),

                        Helpers.Common.CreateAutoAttack(true),
                        Helpers.Common.CreateInterruptBehavior(),

                        Common.CreatePaladinPullMore(),

                        Common.CreatePaladinBlindingLightBehavior(),

                        // Selfless Healer & Bastion of Glory & T16 Free cast
                        Spell.Cast("Flash of Light", 
                            on => StyxWoW.Me, 
                            ret => StyxWoW.Me.GetAuraStacks("Selfless Healer") == 3 && 
                                StyxWoW.Me.HealthPercent <= PaladinSettings.ProtFlashOfLightSHHealth),
                        Spell.Cast(Common.HasTalent(PaladinTalents.EternalFlame) ? "Eternal Flame" : "Word of Glory",
                            on => StyxWoW.Me,
                            ret => StyxWoW.Me.GetAuraStacks("Bastion of Glory") >= (uint)PaladinSettings.ProtBastionOfGloryStacks && 
                                StyxWoW.Me.HealthPercent <= PaladinSettings.ProtBastionOfGloryHealth && 
                                (StyxWoW.Me.CurrentHolyPower >= 3 || (StyxWoW.Me.GetAuraStacks("Bastion of Glory") >= 3 && Has4PieceTier16Bonus))),

                        // Taunts - if reckoning on cooldown, throw some damage at them
                        new Decorator(
                            ret => SingularSettings.Instance.EnableTaunting
                                && TankManager.Instance.NeedToTaunt.Any()
                                && TankManager.Instance.NeedToTaunt.FirstOrDefault().InLineOfSpellSight,
                            new Throttle(TimeSpan.FromMilliseconds(1500),
                                new PrioritySelector(
                                    ctx => TankManager.Instance.NeedToTaunt.FirstOrDefault(e => e.SpellDistance() < 30),
                                    Spell.Cast("Reckoning", ctx => (WoWUnit) ctx),
                                    Spell.Cast("Avenger's Shield", ctx => (WoWUnit)ctx),
                                    Spell.Cast("Judgment", ctx => (WoWUnit)ctx)
                                    )
                                )
                            ),

                        // Soloing move - open with stun to reduce incoming damage (best to take Fist of Justice talent if doing this
                        Spell.Cast("Hammer of Justice", ret => PaladinSettings.StunMobsWhileSolo && SingularRoutine.CurrentWoWContext == WoWContext.Normal),

                        //Multi target
                        new Decorator(
                            ret => _aoeCount >= 3 && Spell.UseAOE,
                            new PrioritySelector(
                                Spell.OffGCD(Spell.Cast("Shield of the Righteous", ret => useSotr())),
                                Spell.Cast("Hammer of the Righteous"),
                                Spell.Cast("Judgment"),
                                Spell.Cast("Holy Wrath", ret => Common.HasTalent(PaladinTalents.SanctifiedWrath)),
                                Spell.Cast("Avenger's Shield"),
                                Spell.Cast("Holy Prism", on => Me),
                                Spell.CastOnGround("Light's Hammer", on => StyxWoW.Me.CurrentTarget, ret => Me.CurrentTarget != null && Clusters.GetClusterCount(Me.CurrentTarget, Unit.NearbyUnfriendlyUnits, ClusterType.Radius, 10f) > 2, false),
                                Spell.Cast("Consecration", ret => !StyxWoW.Me.IsMoving || TalentManager.HasGlyph("the Consecrator")),
                                Spell.Cast("Execution Sentence", on => Unit.NearbyUnfriendlyUnits.Where(u => u.TimeToDeath() > 12 && Me.IsSafelyFacing(u) && u.InLineOfSpellSight).OrderBy(u => u.HealthPercent).FirstOrDefault()),
                                Spell.Cast("Hammer of Wrath", ret => StyxWoW.Me.CurrentTarget.HealthPercent <= 20),
                                Spell.Cast("Holy Wrath"),
                                Movement.CreateMoveToMeleeBehavior(true)
                                )
                            ),

                        //Single target
                        Spell.OffGCD(Spell.Cast("Shield of the Righteous", ret => useSotr())),
                        Spell.Cast("Crusader Strike"),
                        Spell.Cast("Judgment"),
                        Spell.Cast("Holy Wrath", ret => Common.HasTalent(PaladinTalents.SanctifiedWrath)),
                        Spell.Cast("Avenger's Shield"),
                        Spell.Cast("Holy Prism"),
                        Spell.CastOnGround("Light's Hammer", on => StyxWoW.Me.CurrentTarget, ret => StyxWoW.Me.CurrentTarget.DistanceSqr < 10*10, false),
                        Spell.Cast("Execution Sentence", ret => Me.CurrentTarget.TimeToDeath() > 12),
                        Spell.Cast("Hammer of Wrath", ret => Me.CurrentTarget.HealthPercent < 20),
                        Spell.Cast("Consecration", ret => !StyxWoW.Me.IsMoving || TalentManager.HasGlyph("the Consecrator")),
                        Spell.Cast("Holy Wrath")
                        )
                    ),

                Movement.CreateMoveToMeleeBehavior(true));
        }

        private static bool useSotr()
        {
            var HoPo = StyxWoW.Me.CurrentHolyPower;
            bool generatorOnCooldown = Spell.IsSpellOnCooldown("Crusader Strike") && Spell.IsSpellOnCooldown("Hammer of the Righteous") && Spell.IsSpellOnCooldown("Judgement") && Spell.IsSpellOnCooldown("Exorcism") && (Spell.IsSpellOnCooldown("Holy Wrath") && Common.HasTalent(PaladinTalents.SanctifiedWrath) && StyxWoW.Me.HasAura("Avenging Wrath"));

            if (PaladinSettings.SotrHolyPower == SotrPoolHoPo.Never)
            {
                return false;
            }

            if (HoPo == 5 || 
                HoPo >= (uint)PaladinSettings.SotrHolyPower || 
                (HoPo == 4 && !generatorOnCooldown) || 
                StyxWoW.Me.HasAura("Divine Purpose"))
            {
                return true;
            }

            return false;
        }

        private static Composite CreateProtDiagnosticOutputBehavior()
        {
            if (!SingularSettings.Debug)
                return new Action(ret => { return RunStatus.Failure; });

            return new Sequence(
                new ThrottlePasses(1, 1,
                    new Action(ret =>
                    {
                        string sMsg;
                        sMsg = string.Format(".... h={0:F1}%, m={1:F1}%, moving={2}, hpower={3}, grandcru={4}, divpurp={5}, sacshld={6}, mobs={7}",
                            Me.HealthPercent,
                            Me.ManaPercent,
                            Me.IsMoving,
                            Me.CurrentHolyPower,
                            (long)Me.GetAuraTimeLeft("Grand Crusader").TotalMilliseconds,
                            (long)Me.GetAuraTimeLeft("Divine Purpose").TotalMilliseconds,
                            (long)Me.GetAuraTimeLeft("Sacred Shield").TotalMilliseconds, 

                            _aoeCount
                            );

                        WoWUnit target = Me.CurrentTarget;
                        if (target != null)
                        {
                            sMsg += string.Format(
                                ", {0}, {1:F1}%, {2:F1} yds, threat={3}% loss={4}",
                                target.SafeName(),
                                target.HealthPercent,
                                target.Distance,
                                (int) target.ThreatInfo.RawPercent,
                                target.InLineOfSpellSight
                                );
                        }

                        Logger.WriteDebug(Color.LightYellow, sMsg);
                        return RunStatus.Failure;
                    })
                    )
                );
        }
    }
}
