﻿using System.Linq;
using Singular.Dynamics;
using Singular.Helpers;
using Singular.Managers;
using Singular.Settings;
using Styx;

using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Rest = Singular.Helpers.Rest;
using System.Drawing;
using CommonBehaviors.Actions;

namespace Singular.ClassSpecific.Paladin
{
    public class Retribution
    {

        #region Properties & Fields

        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static PaladinSettings PaladinSettings { get { return SingularSettings.Instance.Paladin(); } }

        private const int RET_T13_ITEM_SET_ID = 1064;

        private static int NumTier13Pieces
        {
            get
            {
                return StyxWoW.Me.CarriedItems.Count(i => i.ItemInfo.ItemSetId == RET_T13_ITEM_SET_ID);
            }
        }

        private static bool Has2PieceTier13Bonus { get { return NumTier13Pieces >= 2; } }

        private static int _mobCount;

        #endregion

        #region Heal
        [Behavior(BehaviorType.Heal, WoWClass.Paladin, WoWSpec.PaladinRetribution)]
        public static Composite CreatePaladinRetributionHeal()
        {
            return new PrioritySelector(

                Spell.WaitForCastOrChannel(),

                Spell.BuffSelf("Devotion Aura", req => Me.Silenced),
                Spell.Cast("Lay on Hands",
                    mov => false,
                    on => Me,
                    req => Me.PredictedHealthPercent(includeMyHeals: true) <= PaladinSettings.SelfLayOnHandsHealth),
                Common.CreateWordOfGloryBehavior(on => Me),
                Spell.Cast("Flash of Light",
                    mov => false,
                    on => Me,
                    req => Me.PredictedHealthPercent(includeMyHeals: true) <= PaladinSettings.SelfFlashOfLightHealth,
                    cancel => Me.HealthPercent > PaladinSettings.SelfFlashOfLightHealth)
                );
        }

        [Behavior(BehaviorType.Rest, WoWClass.Paladin, WoWSpec.PaladinRetribution)]
        public static Composite CreatePaladinRetributionRest()
        {
            return new PrioritySelector(
                Spell.WaitForCastOrChannel(),
                new Decorator(
                    ret => !Spell.IsGlobalCooldown(),
                    new PrioritySelector(
                        Rest.CreateDefaultRestBehaviour( "Flash of Light", "Redemption")
                        )
                    )
                );
        }
        #endregion

        #region Normal Rotation

        [Behavior(BehaviorType.Pull | BehaviorType.Combat, WoWClass.Paladin, WoWSpec.PaladinRetribution, WoWContext.Normal | WoWContext.Battlegrounds )]
        public static Composite CreatePaladinRetributionNormalPullAndCombat()
        {
            return CreatePaladinRetributionInstancePullAndCombat();
#if old
            return new PrioritySelector(

                Helpers.Common.EnsureReadyToAttackFromMelee(),

                Spell.WaitForCastOrChannel(),

                new Decorator( 
                    ret => !Spell.IsGlobalCooldown() && Me.GotTarget,
                    new PrioritySelector(
                        // aoe count
                        ActionAoeCount(),

                        CreateRetDiagnosticOutputBehavior(),

                        Helpers.Common.CreateAutoAttack(true),
                        Helpers.Common.CreateInterruptBehavior(),

                        Common.CreatePaladinPullMore(),

                        // Defensive
                        Spell.BuffSelf("Hand of Freedom",
                            ret => Me.HasAuraWithMechanic(WoWSpellMechanic.Dazed,
                                                                  WoWSpellMechanic.Disoriented,
                                                                  WoWSpellMechanic.Frozen,
                                                                  WoWSpellMechanic.Incapacitated,
                                                                  WoWSpellMechanic.Rooted,
                                                                  WoWSpellMechanic.Slowed,
                                                                  WoWSpellMechanic.Snared)),

                        Spell.BuffSelf("Divine Shield", ret => Me.HealthPercent <= 20 && !Me.HasAura("Forbearance") && (!Me.HasAura("Horde Flag") || !Me.HasAura("Alliance Flag"))),
                        Spell.BuffSelf("Divine Protection", ret => Me.HealthPercent <= PaladinSettings.DivineProtectionHealthProt),

                        Common.CreatePaladinSealBehavior(),

                        Spell.Cast( "Hammer of Justice", ret => PaladinSettings.StunMobsWhileSolo && SingularRoutine.CurrentWoWContext == WoWContext.Normal ),

                        //7	Blow buffs seperatly.  No reason for stacking while grinding.
                        Spell.Cast("Guardian of Ancient Kings", ret => PaladinSettings.RetAvengAndGoatK && (_mobCount >= 4 || Me.GotTarget && Me.CurrentTarget.TimeToDeath() > 30)),
                        Spell.Cast("Holy Avenger", ret => PaladinSettings.RetAvengAndGoatK && _mobCount < 4),
                        Spell.BuffSelf("Avenging Wrath", 
                            ret => PaladinSettings.RetAvengAndGoatK
                                && (_mobCount >= 4 || Me.GotTarget && Me.CurrentTarget.TimeToDeath() > 30 || (!Me.HasAura("Holy Avenger") && Spell.GetSpellCooldown("Holy Avenger").TotalSeconds > 10))),

                        Spell.Cast("Execution Sentence", ret => Me.CurrentTarget.TimeToDeath() > 15),
                        Spell.Cast("Holy Prism", on => Group.Tanks.FirstOrDefault(t => t.IsAlive && t.Distance < 40)),

                        new Decorator(
                            ret => _mobCount >= 2 && Spell.UseAOE && Me.CurrentTarget.IsTrivial(),
                            new PrioritySelector(
                                // Bobby53: Inq > 5HP DS > Exo > HotR > 3-4HP DS
                                Spell.Cast("Divine Storm", ret => Me.CurrentHolyPower == 5),
                                Spell.Cast("Exorcism", req => TalentManager.HasGlyph("Mass Exorcism")),
                                Spell.Cast("Hammer of the Righteous"),
                                Spell.Cast("Divine Storm", ret => Me.CurrentHolyPower >= 3)
                                )
                            ),

                        Common.CreatePaladinBlindingLightBehavior(),

                        new Decorator(
                            ret => _mobCount >= 2 && Spell.UseAOE,
                            new PrioritySelector(

                                // was EJ: Inq > 5HP DS > LH > HoW > Exo > HotR > Judge > 3-4HP DS (> SS)
                                // now EJ: Inq > 5HP DS > LH > HoW (> T16 Free DS) > HotR > Judge > Exo > 3-4HP DS (> SS)
                                Spell.Cast(SpellManager.HasSpell("Divine Storm") ? "Divine Storm" : "Templar's Verdict", ret => Me.CurrentHolyPower == 5),
                                Spell.CastOnGround("Light's Hammer", on => Me.CurrentTarget, ret => 2 <= Clusters.GetClusterCount(Me.CurrentTarget, Unit.NearbyUnfriendlyUnits, ClusterType.Radius, 10f)),
                                Spell.Cast("Hammer of Wrath"),
                                Spell.Cast("Divine Storm", req => Me.HasAura("Divine Crusader")),   // T16 buff
                                Spell.Cast(SpellManager.HasSpell("Hammer of the Righteous") ? "Hammer of the Righteous" : "Crusader Strike"),
                                Spell.Cast("Judgment"),
                                Spell.Cast("Exorcism"),
                                Spell.Cast(SpellManager.HasSpell("Divine Storm") ? "Divine Storm" : "Templar's Verdict", ret => Me.CurrentHolyPower >= 3),
                                Spell.BuffSelf("Sacred Shield"),
                                Movement.CreateMoveToMeleeBehavior(true),
                                new ActionAlwaysSucceed()
                                )
                            ),

                        // was EJ: Inq > 5HP TV > ES > HoW > Exo > CS > Judge > 3-4HP TV (> SS)
                        // was EJ: Inq > 5HP TV > ES > HoW > CS > Judge > Exo > 3-4HP TV (> SS)
                        // now EJ: Inq > ES (> 5HP T16 Free DS) > 5HP TV > HoW (> T16 Free DS) > CS > Judge > Exo > 3-4HP TV (> SS)
                        Spell.Cast("Execution Sentence"),
                        Spell.Cast("Divine Storm", req => Me.CurrentHolyPower == 5 && Me.HasAura("Divine Crusader")),   // T16 buff
                        Spell.Cast("Templar's Verdict", req => Me.CurrentHolyPower == 5),
                        Spell.Cast("Hammer of Wrath"),
                        Spell.Cast("Divine Storm", req => Me.HasAura("Divine Crusader")),   // T16 buff
                        Spell.Cast("Crusader Strike"),
                        Spell.Cast("Judgment"),
                        Spell.Cast("Exorcism"),
                        Spell.Cast("Templar's Verdict", req => Me.CurrentHolyPower >= 3),
                        Spell.BuffSelf("Sacred Shield")
                        )
                    ),

                // Move to melee is LAST. Period.
                Movement.CreateMoveToMeleeBehavior(true)
                );
#endif
        }

        private static Action ActionAoeCount()
        {
            return new Action(ret =>
            {
                _mobCount = Unit.NearbyUnfriendlyUnits.Count(u => u.SpellDistance() < 8);
                return RunStatus.Failure;
            });
        }

        #endregion

        #region Instance Rotation

        [Behavior(BehaviorType.Pull | BehaviorType.Combat, WoWClass.Paladin, WoWSpec.PaladinRetribution, WoWContext.Instances)]
        public static Composite CreatePaladinRetributionInstancePullAndCombat()
        {
            return new PrioritySelector(
                Helpers.Common.EnsureReadyToAttackFromMelee(),

                Spell.WaitForCastOrChannel(),

                new Decorator( 
                    ret => !Spell.IsGlobalCooldown() && Me.GotTarget,
                    new PrioritySelector(
                        // aoe count
                        ActionAoeCount(),

                        CreateRetDiagnosticOutputBehavior(),

                        Helpers.Common.CreateAutoAttack(true),

                        Helpers.Common.CreateInterruptBehavior(),

                        #region Defensive & Healing

                        Spell.Cast("Flash of Light", 
                            on => StyxWoW.Me,
                            ret => PaladinSettings.SelfFlashOfLightSH && Me.HealthPercent <= PaladinSettings.SelfFlashOfLightSHHealth && StyxWoW.Me.GetAuraStacks("Selfless Healer") == 3),

                        Spell.Cast("Flash of Light",
                            on => HealerManager.Instance.TargetList.Where(u => u.IsAlive && u.HealthPercent <= PaladinSettings.PartyFlashOfLightSHHealth).OrderBy(u => u.HealthPercent).FirstOrDefault(),
                            req => true),

                        Spell.Cast("Lay on Hands",
                            mov => true,
                            on => HealerManager.Instance.TargetList.Where(u => u.IsAlive && u.HealthPercent <= PaladinSettings.PartyLayOnHandsHealth).OrderBy(u => u.HealthPercent).FirstOrDefault(),
                            req => true),

                        Spell.BuffSelf("Hand of Freedom", ret => 
                            !Me.Auras.Values.Any(a => a.Name.Contains("Hand of") && a.CreatorGuid == Me.Guid) && 
                            Me.HasAuraWithMechanic(WoWSpellMechanic.Dazed,
                                                    WoWSpellMechanic.Disoriented,
                                                    WoWSpellMechanic.Frozen,
                                                    WoWSpellMechanic.Incapacitated,
                                                    WoWSpellMechanic.Rooted,
                                                    WoWSpellMechanic.Slowed,
                                                    WoWSpellMechanic.Snared)),

                        Spell.BuffSelf("Divine Shield", ret => Me.HealthPercent <= PaladinSettings.DivineShieldHealth && !Me.HasAura("Forbearance")),

                        Spell.BuffSelf("Divine Protection", ret => Me.HealthPercent <= PaladinSettings.DivineProtectionHealthProt),

                        #endregion

                        Common.CreatePaladinSealBehavior(),

                        #region Potion during Heroism
                        new Throttle( 40,
                            new Decorator(
                                ret => PartyBuff.WeHaveBloodlust,
                                new PrioritySelector(
                                    ctx => Item.FindFirstUsableItemBySpell("Golem's Strength", "Potion of Mogu Power"),
                                    new Decorator(
                                        ret => ret != null,
                                        new Action(item => ((WoWItem)item).Use() )
                                        )
                                    )
                                )
                            ),
                         #endregion

                        #region Cooldowns

                        new Decorator(
                            ret => Spell.UseCooldowns,
                            new PrioritySelector(
                                Spell.Cast("Avenging Wrath"),
                                Spell.Cast("Holy Avenger")
                                )
                            ),

                        #endregion

                        Spell.Cast( "Execution Sentence", ret => Me.CurrentTarget.TimeToDeath() > 12 ),

                        new Decorator(
                            ret => _mobCount >= 2 && Spell.UseAOE,
                            new PrioritySelector(
                                // T16 Proc
                                Spell.Cast("Divine Storm", ret => StyxWoW.Me.HasAura("Divine Crusader") && (StyxWoW.Me.CurrentHolyPower == 5 || StyxWoW.Me.GetAuraTimeLeft("Divine Crusader").TotalMilliseconds <= 2000)),
                                //Spell.Cast(SpellManager.HasSpell("Divine Storm") ? "Divine Storm" : "Templar's Verdict", ret => StyxWoW.Me.CurrentHolyPower == 5),
                                Spell.Cast(SpellManager.HasSpell("Divine Storm") ? "Divine Storm" : "Templar's Verdict", ret => UseHoPoSpender()),
                                Spell.CastOnGround("Light's Hammer", on => StyxWoW.Me.CurrentTarget, ret => 2 <= Clusters.GetClusterCount(Me.CurrentTarget, Unit.NearbyUnfriendlyUnits, ClusterType.Radius, 10f)),
                                Spell.Cast("Hammer of Wrath"),
                                EmpoweredSeals(),
                                Spell.Cast(SpellManager.HasSpell("Hammer of the Righteous") ? "Hammer of the Righteous" : "Crusader Strike"),
                                Spell.Cast("Judgment", ret => CastJudgment()),
                                // T16 Proc
                                Spell.Cast("Divine Storm", ret => StyxWoW.Me.HasAura("Divine Crusader")),
                                Spell.Cast("Exorcism"),
                                //Spell.Cast(SpellManager.HasSpell("Divine Storm") ? "Divine Storm" : "Templar's Verdict", ret => StyxWoW.Me.CurrentHolyPower >= 3),
                                Spell.Cast("Holy Prism", on => Group.Tanks.FirstOrDefault(t => t.IsAlive && t.Distance < 40)),
                                Spell.BuffSelf("Sacred Shield"),
                                Movement.CreateMoveToMeleeBehavior(true)
                                )
                            ),

                        Spell.Cast("Execution Sentence"),
                        // T16 Proc
                        Spell.Cast("Divine Storm", ret => StyxWoW.Me.HasAura("Divine Crusader") && (StyxWoW.Me.CurrentHolyPower == 5 || StyxWoW.Me.GetAuraTimeLeft("Divine Crusader").TotalMilliseconds <= 2000)),
                        Spell.Cast("Templar's Verdict", ret => UseHoPoSpender()),
                        Spell.Cast("Hammer of Wrath"),
                        EmpoweredSeals(),
                        Spell.Cast("Crusader Strike"),
                        Spell.Cast("Judgment", ret => CastJudgment()),
                        // T16 Proc
                        Spell.Cast("Divine Storm", ret => StyxWoW.Me.HasAura("Divine Crusader")),
                        Spell.Cast("Exorcism"),
                        Spell.Cast("Holy Prism", on => Group.Tanks.FirstOrDefault(t => t.IsAlive && t.Distance < 40)),
                        Spell.Cast("Light's Hammer", on => StyxWoW.Me.CurrentTarget, ret => StyxWoW.Me.CurrentTarget.TimeToDeath() > 12 && !StyxWoW.Me.CurrentTarget.IsMoving),
                        Spell.BuffSelf("Sacred Shield")
                        )
                    ),


                    // Move to melee is LAST. Period.
                    Movement.CreateMoveToMeleeBehavior(true)
                );
        }

        #endregion

        private static Composite CreateRetDiagnosticOutputBehavior()
        {
            if (!SingularSettings.Debug)
                return new Action( ret => { return RunStatus.Failure; } );

            return new Sequence(
                new Action( r => SingularRoutine.UpdateDiagnosticCastingState() ),
                new Action(r => TMsg.ShowTraceMessages = false),
                new ThrottlePasses(1, 1, 
                    new Action(ret =>
                    {
                        TMsg.ShowTraceMessages = true;

                        string sMsg;
                        sMsg = string.Format(".... h={0:F1}%, m={1:F1}%, moving={2}, mobs={3}",
                            Me.HealthPercent,
                            Me.ManaPercent,
                            Me.IsMoving.ToYN(),
                            _mobCount
                            );

                        WoWUnit target = Me.CurrentTarget;
                        if (target != null)
                        {
                            sMsg += string.Format(
                                ", {0}, {1:F1}%, {2:F1} yds, face={3}, loss={4}",
                                target.SafeName(),
                                target.HealthPercent,
                                target.Distance,
                                Me.IsSafelyFacing(target).ToYN(),
                                target.InLineOfSpellSight.ToYN()
                                );
                        }

                        Logger.WriteDebug(Color.LightYellow, sMsg);
                        return RunStatus.Failure;
                    })
                    )
                );
        }

        private static bool UseHoPoSpender()
        {
            var HoPo = StyxWoW.Me.CurrentHolyPower;
            var cooldownMs = 500;
            bool generatorOnCooldown = Spell.GetSpellCooldown("Crusader Strike").TotalMilliseconds > cooldownMs && Spell.GetSpellCooldown("Judgment").TotalMilliseconds > cooldownMs && Spell.GetSpellCooldown("Exorcism").TotalMilliseconds > cooldownMs;

            /*
            Logger.Write("HoPo: {0} | Crusader Strike CD: {1} | Exorcism: {2} | Judgment: {3} | generatorOnCooldown: {4}", 
                HoPo, 
                Spell.GetSpellCooldown("Crusader Strike").TotalMilliseconds, 
                Spell.GetSpellCooldown("Exorcism").TotalMilliseconds, 
                Spell.GetSpellCooldown("Judgment").TotalMilliseconds,
                generatorOnCooldown);
             */

            // Ensure we prioritize getting 5 holy power before using spender
            if (HoPo >= 3 && HoPo != 5 && !generatorOnCooldown)
            {
                return false;
            }

            // We're at 5 Holy Power; Always cast spender
            if (HoPo == 5)
            {
                return true;
            }

            // Holy Avenger Talented
            if (StyxWoW.Me.HasAura("Holy Avenger"))
            {
                if (HoPo <= 2)
                {
                    return false;
                }
                else
                {
                    return true;
                }
            }

            // Divine Purpose Talented
            if ((StyxWoW.Me.HasAura("Divine Purpose") && generatorOnCooldown) || StyxWoW.Me.GetAuraTimeLeft("Divine Purpose").TotalMilliseconds <= 2500)
            {
                return true;
            }

            return false;
        }

        private static Composite EmpoweredSeals()
        {
            return new Decorator(
                ret => StyxWoW.Me.CurrentTarget != null && Common.HasTalent(PaladinTalents.EmpoweredSeals),

                new PrioritySelector(
                    // Make sure we only change back to SoT if Censure is going to fall off, otherwise just stay in SoR until the very last second
                    new Decorator(
                        // Make sure we're always in Seal of Truth if we have SoR and LR (unless we're in aoe?)
                        //ret => StyxWoW.Me.GetAuraTimeLeft("Liadrin's Righteousness").TotalMilliseconds >= 12000 && StyxWoW.Me.HasMyAura("Seal of Righteousness") && (Spell.UseAOE || _mobCount >= 2),
                        ret => StyxWoW.Me.CurrentTarget.GetAuraTimeLeft("Censure").TotalMilliseconds <= 10000 && StyxWoW.Me.HasMyAura("Seal of Righteousness"),
                        Spell.Cast("Seal of Truth")
                        ),
                    // We don't have any buff so just cast Judgment to get one no matter what seal we're in
                    new Decorator(
                        ret => !StyxWoW.Me.HasMyAura("Liadrin's Righteousness") && !StyxWoW.Me.HasMyAura("Maraad's Truth") && Spell.GetSpellCooldown("Judgment").TotalMilliseconds == 0,
                        Spell.Cast("Judgment")
                        ),
                    // SoT Buff
                    new Decorator(
                        ret => (!StyxWoW.Me.HasMyAura("Maraad's Truth") || StyxWoW.Me.GetAuraTimeLeft("Maraad's Truth").TotalMilliseconds <= 5000) && (Spell.GetSpellCooldown("Judgment").TotalMilliseconds <= 900 || StyxWoW.Me.HasMyAura("Seal of Truth")),
                        new PrioritySelector(
                            Spell.Cast("Seal of Truth", ret => !StyxWoW.Me.HasMyAura("Seal of Truth")),
                            Spell.Cast("Judgment")
                            )
                        ),
                    // SoR Buff
                    new Decorator(
                        ret => (!StyxWoW.Me.HasMyAura("Liadrin's Righteousness") || StyxWoW.Me.GetAuraTimeLeft("Liadrin's Righteousness").TotalMilliseconds <= 5000) && (Spell.GetSpellCooldown("Judgment").TotalMilliseconds <= 900 || StyxWoW.Me.HasMyAura("Seal of Righteousness")),
                        new PrioritySelector(
                            Spell.Cast("Seal of Righteousness", ret => !StyxWoW.Me.HasMyAura("Seal of Righteousness")),
                            Spell.Cast("Judgment")
                            )
                        )
                    )
                );
        }

        private static bool CastJudgment()
        {
            if (Common.HasTalent(PaladinTalents.EmpoweredSeals) && (StyxWoW.Me.GetAuraTimeLeft("Liadrin's Righteousness").TotalMilliseconds <= 5000 || StyxWoW.Me.GetAuraTimeLeft("Maraad's Truth").TotalMilliseconds <= 5000))
            {
                return false;
            }

            return true;
        }

    }
    
    public class TMsg : Decorator
    {
        public static bool ShowTraceMessages { get; set; }

        public TMsg(SimpleStringDelegate str)
            : base(ret => ShowTraceMessages, new Action(r => { Logger.WriteDebug(str(r)); return RunStatus.Failure; }))
        {
        }
    }

}
