﻿
using System;
using System.ComponentModel;
using System.IO;
using Singular.ClassSpecific.Paladin;

using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Styx.CommonBot;
using Styx;

namespace Singular.Settings
{

    public enum PaladinBlessings
    {
        Auto,
        Kings,
        Might
    }

    public enum PaladinSeal
    {
        None = 0,
        Auto = 1,
        Command,
        Truth,
        Insight,
        Righteousness,
        Justice
    }

    public enum SotrPoolHoPo
    {
        Never,
        _3 = 3,
        _4 = 4,
        _5 = 5
    }

    public enum BastionOfGloryStacks
    {
        _0 = 0,
        _1 = 1,
        _2 = 2,
        _3 = 3,
        _4 = 4,
        _5 = 5
    }

    public enum WordOfGloryHoPo
    {
        _0 = 0,
        _1 = 1,
        _2 = 2,
        _3 = 3,
        _4 = 4,
        _5 = 5
    }

    public enum LightOfDawnHoPo
    {
        _0 = 0,
        _1 = 1,
        _2 = 2,
        _3 = 3,
        _4 = 4,
        _5 = 5
    }

    internal class PaladinSettings : Styx.Helpers.Settings
    {
        public PaladinSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Paladin.xml"))
        {
        }


        #region Common
        /*
        [Setting]
        [DefaultValue(PaladinAura.Auto)]
        [Category("Common")]
        [DisplayName("Aura")]
        [Description("The aura to be used while not mounted. Set this to Auto to allow the CC to automatically pick the aura depending on spec.")]
        public PaladinAura Aura { get; set; }
        */

        [Setting]
        [DefaultValue(PaladinBlessings.Auto)]
        [Category("Common")]
        [DisplayName("Blessings")]
        [Description("Which Blessing to cast (Auto: best choice)")]
        public PaladinBlessings Blessings { get; set; }

        [Setting]
        [DefaultValue(PaladinSeal.Auto)]
        [Category("Common")]
        [DisplayName("Seal")]
        [Description("Which Seal to cast (None: user controlled, Auto: best choice)")]
        public PaladinSeal Seal { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Common")]
        [DisplayName("Hammer of Justice immediately in Normal Context")]
        [Description("Stun mobs while Solo immediately to reduce damage taken")]
        public bool StunMobsWhileSolo { get; set; }

        #endregion

        #region Common Cooldowns

        [Setting]
        [DefaultValue(35)]
        [Category("Common Cooldowns")]
        [DisplayName("Hand of Sacrifice")]
        [Description("Hand of Sacrifice will be used at this value")]
        public int HandOfSacrificeHealth { get; set; }

        #endregion

        #region Self-Heal

        [Setting]
        [DefaultValue(15)]
        [Category("Self-Healing")]
        [DisplayName("Divine Shield Health %")]
        [Description("Divine Shield will be used at this value")]
        public int DivineShieldHealth { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("Self-Healing")]
        [DisplayName("Lay on Hand Health %")]
        [Description("Lay on Hands will be used at this value")]
        public int SelfLayOnHandsHealth { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Self-Healing")]
        [DisplayName("Flash of Light - Selfless Healer")]
        [Description("True: enable using FoL with SH - False: disable using FoL with SH")]
        public bool SelfFlashOfLightSH { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("Self-Healing")]
        [DisplayName("Flash of Light - Selfless Healer Self Health %")]
        [Description("Flash of Light @ 3 stacks Selfless Healer will be used at this value")]
        public int SelfFlashOfLightSHHealth { get; set; }  

        [Setting]
        [DefaultValue(40)]
        [Category("Self-Healing")]
        [DisplayName("Flash of Light Health %")]
        [Description("Flash of Light will be used at this value")]
        public int SelfFlashOfLightHealth { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("Self-Healing")]
        [DisplayName("Word of Glory Health % (1+ Holy Power)")]
        [Description("Health % to Cast Word of Glory if only 1+ Holy Power")]
        public int SelfWordOfGloryHealth1 { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("Self-Healing")]
        [DisplayName("Word of Glory Health % (2+ Holy Power)")]
        [Description("Health % to Cast Word of Glory at 2+ Holy Power")]
        public int SelfWordOfGloryHealth2 { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("Self-Healing")]
        [DisplayName("Word of Glory Health % (3+ Holy Power)")]
        [Description("Health % to Cast Word of Glory at 3+ Holy Power")]
        public int SelfWordOfGloryHealth3 { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("Self-Healing")]
        [DisplayName("Eternal Flame Health %")]
        [Description("Health % to cast Eternal Flame at 3+ Holy Power")]
        public int SelfEternalFlameHealth { get; set; }

        #endregion  

        #region Holy

        [Setting]
        [DefaultValue(false)]
        [Category("Holy")]
        [DisplayName("Beacon of Light Tank Only")]
        [Description("Beacon of Light will only be used on Tank if set to True")]
        public bool BeaconOfLightTankOnly { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("Holy")]
        [DisplayName("Beacon of Light Min Tank Health %")]
        [Description("Beacon of Light will always be placed on tank at or below this value")]
        public int BeaconOfLightTankHealth { get; set; }

        [Setting]
        [DefaultValue(45)]
        [Category("Holy")]
        [DisplayName("Beacon of Light Min Health %")]
        [Description("Any player will be targeted by Beacon of Light if below this value (value does nothing if Beacon Of Light Tank Only is set to true)")]
        public int BeaconOfLightAnyHealth { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("Holy")]
        [DisplayName("Crusader Strike HP Generator")]
        [Description("Use Crusader Strike if in melee distance and targeting enemy")]
        public bool UseCrusaderStrike { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("Holy")]
        [DisplayName("Flash of Light Health %")]
        [Description("Flash of Light will be used at this value")]
        public int FlashOfLightHealth { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("Holy")]
        [DisplayName("Flash of Light Tank Emergency Health %")]
        [Description("Flash of Light will be used at this value on Tanks and will be given higher priority")]
        public int FlashOfLightTankHealth { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("Holy")]
        [DisplayName("Flash of Light Anyone Emergency Health %")]
        [Description("Flash of Light will be used at this value on Anyone and will be given higher priority (less than FoL Tank emergency)")]
        public int FlashOfLightAnyoneHealth { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Holy")]
        [DisplayName("Eternal Flame on Tank priority")]
        [Description("Ensure that Eternal Flame remains on the tank.")]
        public bool KeepEternalFlameUp { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("Holy")]
        [DisplayName("Eternal Flame # Holy Power Tank")]
        [Description("How many holy power to use on Eternal Flame for Tanks")]
        public int EternalFlameHolyPowerTank { get; set; }

        [Setting]
        [DefaultValue(85)]
        [Category("Holy")]
        [DisplayName("Eternal Flame Health %")]
        [Description("Health % to cast Eternal Flame")]
        public int EternalFlameHealth { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("Holy")]
        [DisplayName("Eternal Flame # Holy Power")]
        [Description("How many holy power to use on Eternal Flame")]
        public int EternalFlameHolyPower { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("Common Cooldowns")]
        [DisplayName("Lay on Hand Health %")]
        [Description("Lay on Hands will be used at this value")]
        public int LayOnHandsHealth { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("Holy")]
        [DisplayName("Light of Dawn Health")]
        [Description("Light of Dawn will be used at this value")]
        public int LightOfDawnHealth { get; set; }

        [Setting]
        [DefaultValue(6)]
        [Category("Holy")]
        [DisplayName("Light of Dawn Count")]
        [Description("Light of Dawn will be used when there are more then that many players with lower health then LoD Health setting")]
        public int LightOfDawnCount { get; set; }

        [Setting]
        [DefaultValue(LightOfDawnHoPo._3)]
        [Category("Holy")]
        [DisplayName("Light of Dawn Minimum Holy Power")]
        [Description("Light of Dawn will be used when Holy Power is equal to or greater than this value")]
        public LightOfDawnHoPo LightOfDawnHP { get; set; }

        [Setting]
        [DefaultValue(100)]
        [Category("Holy")]
        [DisplayName("Holy Shock Health")]
        [Description("Holy Shock will be used at this value")]
        public int HolyShockHealth { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("Holy")]
        [DisplayName("Holy Light Health")]
        [Description("Holy Light will be used at this value")]
        public int HolyLightHealth { get; set; }

        [Setting]
        [DefaultValue(85)]
        [Category("Holy")]
        [DisplayName("Holy Radiance Health")]
        [Description("Holy Radiance will be used at this value")]
        public int HolyRadianceHealth { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("Holy")]
        [DisplayName("Holy Radiance Count")]
        [Description("Holy Radiance will be used when there are more then this many players with lower health then HR Health setting")]
        public int HolyRadianceCount { get; set; }

        [Setting]
        [DefaultValue(45)]
        [Category("Holy")]
        [DisplayName("Word of Glory Health")]
        [Description("Word of Glory will be used at this value")]
        public int WordOfGloryHealth { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("Holy")]
        [DisplayName("Word of Glory Emergency Health")]
        [Description("Word of Glory will be used for emergency healing at this value")]
        public int WordOfGloryEmergencyHealth { get; set; }

        [Setting]
        [DefaultValue(WordOfGloryHoPo._3)]
        [Category("Holy")]
        [DisplayName("Word of Glory Minimum Holy Power")]
        [Description("Word of Glory will be used when Holy Power is equal to or greater than this value")]
        public WordOfGloryHoPo WordOfGloryHP { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Holy")]
        [DisplayName("Sacred Shield Tank Priority")]
        [Description("True: sacred shield will always prioritize tanks - False: sacred shield will not prioritize tank")]
        public bool SacredShieldTankPriority { get; set; }

        [Setting]
        [DefaultValue(45)]
        [Category("Holy")]
        [DisplayName("Sacred Shield Health")]
        [Description("Sacred Shield will be used at this value")]
        public int SacredShieldHealth { get; set; }

         #endregion

        #region Protection
        [Setting]
        [DefaultValue(SotrPoolHoPo._3)]
        [Category("Protection")]
        [DisplayName("Shield of the Righteous Holy Power")]
        [Description("Guardian of Ancient Kings will be used when Holy Power >= this value")]
        public SotrPoolHoPo SotrHolyPower { get; set; }
        
        [Setting]
        [DefaultValue(40)]
        [Category("Protection")]
        [DisplayName("Guardian of Ancient Kings Health")]
        [Description("Guardian of Ancient Kings will be used at this value")]
        public int GoAKHealth { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("Protection")]
        [DisplayName("Selfless Healer Flash of Light Health")]
        [Description("Flash of Light with Selfless Healer at 3 Stacks will be used on myself at this value")]
        public int ProtFlashOfLightSHHealth { get; set; }

        [Setting]
        [DefaultValue(BastionOfGloryStacks._5)]
        [Category("Protection")]
        [DisplayName("Bastion of Glory Stacks")]
        [Description("Min amount of Bastion of Glory stacks to use EF/Wog")]
        public BastionOfGloryStacks ProtBastionOfGloryStacks { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("Protection")]
        [DisplayName("Bastion of Glory EF/WoG Health")]
        [Description("EF/Wog will be used on myself at this value with the specified number of Bastion of Glory stacks. EF/WoG will always be used @ 3 or more holy power")]
        public int ProtBastionOfGloryHealth { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("Protection")]
        [DisplayName("Ardent Defender Health")]
        [Description("Ardent Defender will be used at this value")]
        public int ArdentDefenderHealth { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("Protection")]
        [DisplayName("Divine Protection Health")]
        [Description("Divine Protection will be used at this value")]
        public int DivineProtectionHealthProt { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("Protection")]
        [DisplayName("Avengers On Pull Only")]
        [Description("Only use Avenger's Shield to pull")]
        public bool AvengersPullOnly { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("Protection")]
        [DisplayName("Consecration Count")]
        [Description("Consecration will be used when you have more then that many mobs attacking you")]
        public int ProtConsecrationCount { get; set; }
        #endregion

        #region Retribution
        [Setting]
        [DefaultValue(70)]
        [Category("Retribution")]
        [DisplayName("Divine Protection Health")]
        [Description("Divine Protection will be used at this value")]
        public int DivineProtectionHealthRet { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Retribution")]
        [DisplayName("Use Aveng Wrath/Holy Aveng/GotAK")]
        [Description("True: Automatically use Avenging Wrath, Holy Avenger, and Guardian of the Ancient Kings.  False: you will have to cast manually.")]
        public bool RetAvengAndGoatK { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("Retribution")]
        [DisplayName("Flash of Light - Selfless Healer Party Health %")]
        [Description("Flash of Light @ 3 stacks Selfless Healer will be used at this value")]
        public int PartyFlashOfLightSHHealth { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("Retribution")]
        [DisplayName("Lay on Hand Health %")]
        [Description("Lay on Hands will be used on anyone at this value")]
        public int PartyLayOnHandsHealth { get; set; }

       #endregion
    }
}